import logging
from typing import Any, Dict, List, Optional, Type

import httpx
from httpx import AsyncClient
from pydantic import BaseModel

from .models.client_config_models import TicketOperation, OTOBOClientConfig
from .models.request_models import TicketSearchParams, TicketCreateParams, TicketHistoryParams, TicketUpdateParams, \
    TicketGetParams
from .models.response_models import OTOBOTicketCreateResponse, OTOBOTicketGetResponse, \
    OTOBOTicketHistoryResponse, TicketUpdateResponse, TicketSearchResponse, FullTicketSearchResponse, TicketGetResponse
from .otobo_errors import OTOBOError
from .util.http_method import HttpMethod


class OTOBOClient:
    def __init__(self, config: OTOBOClientConfig, client: httpx.AsyncClient = None):
        self._client = client or AsyncClient()
        self.config = config
        self.base_url = config.base_url.rstrip('/')
        self.service = config.service
        self.auth = config.auth
        self._logger = logging.getLogger(__name__)

    def _build_url(self, endpoint: str) -> str:
        """Build the full URL for the OTOBO service."""
        return f"{self.base_url}/Webservice/{self.service}/{endpoint}"

    def _check_operation_registered(self, op_key: TicketOperation) -> None:
        """Check if the operation is registered in the client configuration."""
        if op_key not in self.config.operations:
            raise RuntimeError(f"Operation '{op_key}' is not configured in OTOBOClientConfig")

    def _check_response(self, response: Dict[str, Any]) -> None:
        """Check if the response contains an error."""
        if "Error" in response:
            err = response["Error"]
            raise OTOBOError(err["ErrorCode"], err["ErrorMessage"])

    async def _call[T: BaseModel](self,
                                  method: HttpMethod,
                                  op_key: TicketOperation,
                                  response_model: Type[T],
                                  data: Optional[Dict[str, Any]] = None,
                                  ) -> T:
        self._check_operation_registered(op_key)
        url = self._build_url(self.config.operations[op_key])
        headers = {'Content-Type': 'application/json'}
        payload: Dict[str, Any] = self.auth.model_dump(exclude_none=True) | data or {}
        resp = await self._client.request(method.value, url, json=payload, headers=headers)
        json_response = resp.json()
        self._check_response(json_response)
        resp.raise_for_status()
        return response_model.model_validate(json_response)

    async def create_ticket(self, payload: TicketCreateParams) -> OTOBOTicketCreateResponse:
        return await self._call(HttpMethod.POST, TicketOperation.CREATE, OTOBOTicketCreateResponse,
                                data=payload.model_dump(exclude_none=True))

    async def get_ticket(self, params: TicketGetParams) -> TicketGetResponse:
        otobo_ticket_get_response = await self._call(HttpMethod.POST, TicketOperation.GET, OTOBOTicketGetResponse,
                                                     data=params.model_dump(exclude_none=True))
        tickets = otobo_ticket_get_response.Ticket
        assert len(tickets) == 1, "Expected exactly one ticket in the response"
        return TicketGetResponse(Ticket=tickets[0])

    async def update_ticket(self, payload: TicketUpdateParams) -> TicketUpdateResponse:
        return await self._call(HttpMethod.PUT, TicketOperation.UPDATE, TicketUpdateResponse,
                                data=payload.model_dump(exclude_none=True))

    async def search_tickets(self, query: TicketSearchParams) -> TicketSearchResponse:
        return await self._call(HttpMethod.POST, TicketOperation.SEARCH, TicketSearchResponse,
                                data=query.model_dump(exclude_none=True))

    async def get_ticket_history(self, payload: TicketHistoryParams) -> OTOBOTicketHistoryResponse:
        return await self._call(HttpMethod.POST, TicketOperation.HISTORY_GET, OTOBOTicketHistoryResponse,
                                data=payload.model_dump(exclude_none=True))

    async def search_and_get(self, query: TicketSearchParams) -> FullTicketSearchResponse:
        if TicketOperation.SEARCH not in self.config.operations or TicketOperation.GET not in self.config.operations:
            raise RuntimeError("Both 'TicketSearch' and 'TicketGet' must be configured for search_and_get")
        ids = (await self.search_tickets(query)).TicketID
        ticket_get_responses: List[OTOBOTicketGetResponse] = [await self.get_ticket(TicketGetParams(TicketID=i)) for i
                                                              in ids]
        return [ticket.Ticket for ticket in ticket_get_responses]



from typing import Any, Dict

import httpx
import pytest

from models.client_config_models import TicketOperation, OTOBOClientConfig
from models.request_models import (
    TicketCreateParams, TicketGetParams, TicketUpdateParams,
    TicketSearchParams
)
from models.response_models import (
    OTOBOTicketCreateResponse, OTOBOTicketGetResponse,
    TicketUpdateResponse, TicketSearchResponse, TicketGetResponse
)
from models.ticket_models import TicketDetailOutput, ArticleDetail
from otobo_client import OTOBOClient
from otobo_errors import OTOBOError


# Dummy Async Client to inject
class DummyAsyncClient:
    def __init__(self, response_json: Dict[str, Any], status: int = 200):
        self._response_json = response_json
        self._status = status

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        return False

    async def request(self, method: str, url: str, json=None, headers=None):
        class DummyResponse:
            def __init__(self, data, status):
                self._data = data
                self.status_code = status

            def json(self):
                return self._data

            def raise_for_status(self):
                if self.status_code >= 400:
                    raise httpx.HTTPStatusError('Error', request=None, response=None)

        return DummyResponse(self._response_json, self._status)


@pytest.fixture
def base_config():
    auth = {
        'UserLogin': 'u', 'Password': 'p'
    }
    return OTOBOClientConfig(
        base_url='http://example.com/otobo',
        service='SVC',
        auth=auth,
        operations={
            TicketOperation.CREATE: 'TicketCreate',
            TicketOperation.GET: 'TicketGet',
            TicketOperation.UPDATE: 'TicketUpdate',
            TicketOperation.SEARCH: 'TicketSearch',
            TicketOperation.HISTORY_GET: 'TicketHistoryGet',
        }
    )


@pytest.mark.asyncio
async def test_create_ticket_success(base_config):
    dummy = DummyAsyncClient(
        OTOBOTicketCreateResponse(
            TicketNumber="2039",
            TicketID=1,
            Ticket=TicketDetailOutput(
                Article=ArticleDetail(),
                DynamicField=[]
            ),
            ArticleID=1)
    )
    client = OTOBOClient(base_config, client=dummy)
    payload = TicketCreateParams(
        Ticket={'Title': 'T', 'Queue': 'Q', 'State': 'new', 'Priority': '3 normal', 'CustomerUser': 'c'},
        Article={'Subject': 's', 'Body': 'b', 'MimeType': 'text/plain'})
    res = await client.create_ticket(payload)
    assert isinstance(res, OTOBOTicketCreateResponse)
    assert res.TicketID == 1


@pytest.mark.asyncio
async def test_get_ticket_success(base_config):
    data = OTOBOTicketGetResponse(
        Ticket=[TicketDetailOutput(TicketID=5, Title='Test Ticket', Article=ArticleDetail(), DynamicField=[])])
    dummy = DummyAsyncClient(data)
    client = OTOBOClient(base_config, client=dummy)
    params = TicketGetParams(TicketID=5)
    res = await client.get_ticket(params)
    assert isinstance(res, TicketGetResponse)
    assert res.Ticket.TicketID == 5


@pytest.mark.asyncio
async def test_update_ticket_success(base_config):
    data = TicketUpdateResponse(TicketID=7, Ticket=TicketDetailOutput(Article=ArticleDetail(), DynamicField=[]))
    dummy = DummyAsyncClient(data)
    client = OTOBOClient(base_config, client=dummy)
    payload = TicketUpdateParams(TicketID=7)
    res = await client.update_ticket(payload)
    assert isinstance(res, TicketUpdateResponse)
    assert res.TicketID == 7


@pytest.mark.asyncio
async def test_search_tickets_success(base_config):
    data = {'TicketID': [10, 11]}
    dummy = DummyAsyncClient(data)
    client = OTOBOClient(base_config, client=dummy)
    query = TicketSearchParams()
    res = await client.search_tickets(query)
    assert isinstance(res, TicketSearchResponse)
    assert res.TicketID == [10, 11]


@pytest.mark.asyncio
async def test_error_response_raises(base_config):
    error = {'Error': {'ErrorCode': 'X', 'ErrorMessage': 'msg'}}
    dummy = DummyAsyncClient(error)
    client = OTOBOClient(base_config, client=dummy)
    with pytest.raises(OTOBOError) as ei:
        await client.create_ticket(
            TicketCreateParams(
                Ticket={'Title': 'T', 'Queue': 'Q', 'State': 'new', 'Priority': '3 normal', 'CustomerUser': 'c'},
                Article={'Subject': 's', 'Body': 'b', 'MimeType': 'text/plain'})
        )
    assert 'X' in str(ei.value)


@pytest.fixture
async def client_with_dummy(base_config):
    # default dummy returns empty dict
    dummy = DummyAsyncClient({})
    return OTOBOClient(base_config, client=dummy)



@pytest.mark.asyncio
async def test_search_tickets_success(base_config):
    response_json = {'TicketID': [10, 11]}
    dummy = DummyAsyncClient(response_json)
    client = OTOBOClient(base_config, client=dummy)
    query = TicketSearchParams()
    res = await client.search_tickets(query)
    assert isinstance(res, TicketSearchResponse)
    assert res.TicketID == [10, 11]




@pytest.mark.asyncio
async def test_operation_not_registered(base_config):
    bad_config = base_config.copy(deep=True)
    bad_config.operations.pop(TicketOperation.CREATE)
    dummy = DummyAsyncClient({'any': 'thing'})
    client = OTOBOClient(bad_config, client=dummy)
    payload = TicketCreateParams(
        Ticket={'Title': 'T', 'Queue': 'Q', 'State': 'new', 'Priority': '3 normal', 'CustomerUser': 'c'},
        Article={'Subject': 's', 'Body': 'b', 'MimeType': 'text/plain'}
    )
    with pytest.raises(RuntimeError):
        await client.create_ticket(payload)


@pytest.mark.asyncio
async def test_error_response_raises(base_config):
    error_json = {'Error': {'ErrorCode': 'X', 'ErrorMessage': 'msg'}}
    dummy = DummyAsyncClient(error_json)
    client = OTOBOClient(base_config, client=dummy)
    with pytest.raises(OTOBOError) as exc:
        await client.create_ticket(
            TicketCreateParams(
                Ticket={'Title': 'T', 'Queue': 'Q', 'State': 'new', 'Priority': '3 normal', 'CustomerUser': 'c'},
                Article={'Subject': 's', 'Body': 'b', 'MimeType': 'text/plain'}
            )
        )
    assert exc.value.code == 'X'
    assert exc.value.message == 'msg'



from typing import Union, List, Optional

from pydantic import BaseModel

from .ticket_models import TicketDetailOutput, TicketCommon, ArticleDetail


class OTOBOTicketCreateResponse(BaseModel):
    TicketNumber: str
    TicketID: Union[int, str]
    Ticket: TicketDetailOutput
    ArticleID: int


# --- Models for Ticket Get Response ---
class OTOBOTicketGetResponse(BaseModel):
    Ticket: List[TicketDetailOutput]


class TicketGetResponse(BaseModel):
    Ticket: TicketDetailOutput


# --- Models for Ticket TicketHistoryModel Response ---
class OTOBOTicketHistoryEntry(TicketCommon):
    HistoryType: str
    HistoryTypeID: int
    Name: str



class TicketHistoryModel(BaseModel):
    TicketID: int
    History: List[OTOBOTicketHistoryEntry] = []


class OTOBOTicketHistoryResponse(BaseModel):
    TicketHistory: List[TicketHistoryModel]


class TicketUpdateResponse(BaseModel):
    TicketID: int
    ArticleID: Optional[int] = None
    Ticket: TicketDetailOutput


class TicketSearchResponse(BaseModel):
    TicketID: List[int]


class FullTicketSearchResponse(BaseModel):
    Ticket: List[TicketDetailOutput] = []


from enum import Enum
from typing import Dict

from pydantic import BaseModel, Field

from .request_models import AuthData


class TicketOperation(Enum):
    CREATE = "TicketCreate"
    SEARCH = "TicketSearch"
    GET = "TicketGet"
    UPDATE = "TicketUpdate"
    HISTORY_GET = "TicketHistoryGet"


class OTOBOClientConfig(BaseModel):
    base_url: str = Field(...,
                          description="Base URL of the OTOBO installation, e.g. https://server/otobo/nph-genericinterface.pl")
    service: str = Field(..., description="Webservice connector name")
    auth: AuthData
    operations: Dict[TicketOperation, str] = Field(...,
                                                   description="Mapping of operation keys to endpoint names, e.g. {'TicketCreate': 'ticket-create', ...}")





from typing import Optional, Union, List, Dict, Literal

from pydantic import BaseModel, Field

from .ticket_models import TicketDetailInput


class AuthData(BaseModel):
    SessionID: Optional[int] = None
    UserLogin: str = Field(..., description="Agent login for authentication")
    Password: str = Field(..., description="Agent password for authentication")


class DynamicFieldParams(BaseModel):
    Empty: Optional[bool] = None
    Equals: Optional[Union[int, str, List[Union[int, str]]]] = None
    Like: Optional[str] = None
    GreaterThan: Optional[str] = None
    GreaterThanEquals: Optional[str] = None
    SmallerThan: Optional[str] = None
    SmallerThanEquals: Optional[str] = None


class TicketSearchParams(BaseModel):
    TicketNumber: Optional[Union[str, List[str]]] = None
    Title: Optional[Union[str, List[str]]] = None
    Queues: Optional[List[str]] = None
    QueueIDs: Optional[List[int]] = None
    UseSubQueues: Optional[bool] = None
    Types: Optional[List[str]] = None
    TypeIDs: Optional[List[int]] = None
    States: Optional[List[str]] = None
    StateIDs: Optional[List[int]] = None
    StateType: Optional[Union[str, List[str]]] = None
    StateTypeIDs: Optional[List[int]] = None
    Priorities: Optional[List[str]] = None
    PriorityIDs: Optional[List[int]] = None
    Services: Optional[List[str]] = None
    ServiceIDs: Optional[List[int]] = None
    SLAs: Optional[List[str]] = None
    SLAIDs: Optional[List[int]] = None
    Locks: Optional[List[str]] = None
    LockIDs: Optional[List[int]] = None
    OwnerIDs: Optional[List[int]] = None
    ResponsibleIDs: Optional[List[int]] = None
    WatchUserIDs: Optional[List[int]] = None
    CustomerID: Optional[Union[str, List[str]]] = None
    CustomerIDRaw: Optional[Union[str, List[str]]] = None
    CustomerUserLogin: Optional[Union[str, List[str]]] = None
    CreatedUserIDs: Optional[List[int]] = None
    CreatedTypes: Optional[List[str]] = None
    CreatedTypeIDs: Optional[List[int]] = None
    CreatedPriorities: Optional[List[str]] = None
    CreatedPriorityIDs: Optional[List[int]] = None
    CreatedStates: Optional[List[str]] = None
    CreatedStateIDs: Optional[List[int]] = None
    CreatedQueues: Optional[List[str]] = None
    CreatedQueueIDs: Optional[List[int]] = None
    DynamicFields: Optional[Dict[str, DynamicFieldParams]] = None
    MIMEBase_From: Optional[str] = None
    MIMEBase_To: Optional[str] = None
    MIMEBase_Cc: Optional[str] = None
    MIMEBase_Subject: Optional[str] = None
    MIMEBase_Body: Optional[str] = None
    AttachmentName: Optional[str] = None
    FullTextIndex: Optional[bool] = None
    ContentSearch: Optional[str] = None
    ConditionInline: Optional[bool] = None
    ArticleCreateTimeOlderMinutes: Optional[int] = None
    ArticleCreateTimeNewerMinutes: Optional[int] = None
    ArticleCreateTimeNewerDate: Optional[str] = None
    ArticleCreateTimeOlderDate: Optional[str] = None
    TicketCreateTimeOlderMinutes: Optional[int] = None
    TicketCreateTimeNewerMinutes: Optional[int] = None
    TicketCreateTimeNewerDate: Optional[str] = None
    TicketCreateTimeOlderDate: Optional[str] = None
    TicketChangeTimeOlderMinutes: Optional[int] = None
    TicketChangeTimeNewerMinutes: Optional[int] = None
    TicketLastChangeTimeOlderMinutes: Optional[int] = None
    TicketLastChangeTimeNewerMinutes: Optional[int] = None
    TicketLastChangeTimeNewerDate: Optional[str] = None
    TicketLastChangeTimeOlderDate: Optional[str] = None
    TicketChangeTimeNewerDate: Optional[str] = None
    TicketChangeTimeOlderDate: Optional[str] = None
    TicketCloseTimeOlderMinutes: Optional[int] = None
    TicketCloseTimeNewerMinutes: Optional[int] = None
    TicketCloseTimeNewerDate: Optional[str] = None
    TicketCloseTimeOlderDate: Optional[str] = None
    TicketPendingTimeOlderMinutes: Optional[int] = None
    TicketPendingTimeNewerMinutes: Optional[int] = None
    TicketPendingTimeNewerDate: Optional[str] = None
    TicketPendingTimeOlderDate: Optional[str] = None
    TicketEscalationTimeOlderMinutes: Optional[int] = None
    TicketEscalationTimeNewerMinutes: Optional[int] = None
    TicketEscalationTimeNewerDate: Optional[str] = None
    TicketEscalationTimeOlderDate: Optional[str] = None
    SearchInArchive: Optional[str] = None
    OrderBy: Optional[Union[str, List[str]]] = None
    SortBy: Optional[Union[str, List[str]]] = None

    def model_dump(self, **kwargs):
        data = super().model_dump(**kwargs)
        df = data.pop("DynamicFields", {}) or {}
        for name, params in df.items():
            data[f"DynamicField_{name}"] = params
        return data


class TicketCreateParams(TicketDetailInput):
    pass


class TicketGetParams(BaseModel):
    TicketID: Optional[int] = None
    DynamicFields: int = 1
    Extended: int = 1
    AllArticles: int = 1
    ArticleSenderType: Optional[List[str]] = None
    ArticleOrder: Literal["ASC", "DESC"] = 'ASC'
    ArticleLimit: int = 20
    Attachments: int = 0
    GetAttachmentContents: int = 1
    HTMLBodyAsAttachment: int = 1


class TicketHistoryParams(BaseModel):
    TicketID: str


class TicketUpdateParams(TicketDetailInput):
    TicketID: Optional[int] = None
    TicketNumber: Optional[str] = None



from typing import Any, List, Union
from typing import Optional

from pydantic import BaseModel



class AttachmentModel(BaseModel):
    Filename: str
    ContentType: Optional[str] = None
    Content: Optional[str] = None  # Base64 encoded content


class AttachmentResponseDetail(AttachmentModel):
    ContentAlternative: Optional[str] = None
    ContentID: Optional[str] = None
    Filesize: Optional[str] = None
    FilesizeRaw: Optional[int] = None


class TicketBase(BaseModel):
    Title: Optional[str] = None
    QueueID: Optional[int] = None
    Queue: Optional[str] = None
    LockID: Optional[int] = None
    Lock: Optional[str] = None
    TypeID: Optional[int] = None
    Type: Optional[str] = None
    ServiceID: Optional[Union[int, str]]= None
    Service: Optional[str] = None
    SLAID: Optional[Union[int, str]] = None
    SLA: Optional[str] = None
    StateID: Optional[int] = None
    State: Optional[str] = None
    PriorityID: Optional[int] = None
    Priority: Optional[str] = None
    OwnerID: Optional[int] = None
    Owner: Optional[str] = None
    ResponsibleID: Optional[int] = None
    Responsible: Optional[str] = None
    CustomerUser: Optional[str] = None


class TicketCommon(TicketBase):
    TicketID: Optional[int] = None
    TicketNumber: Optional[str] = None
    StateType: Optional[str] = None
    CustomerID: Optional[str] = None
    CustomerUserID: Optional[str] = None
    ArchiveFlag: Optional[str] = None
    Age: Optional[int] = None
    EscalationResponseTime: Optional[int] = None
    EscalationUpdateTime: Optional[int] = None
    EscalationSolutionTime: Optional[int] = None
    EscalationTime: Optional[int] = None
    CreateBy: Optional[int] = None
    ChangeBy: Optional[int] = None
    Created: Optional[str] = None
    Changed: Optional[str] = None


# --- Models for Ticket Create Response ---
class DynamicFieldItem(BaseModel):
    Name: str
    Value: Optional[Any] = None


from typing import Optional, Union
from pydantic import BaseModel


class ArticleDetail(BaseModel):
    CommunicationChannel: Optional[str] = None
    CommunicationChannelID: Optional[int] = None
    IsVisibleForCustomer: Optional[int] = None
    SenderTypeID: Optional[Union[int, str]] = None
    AutoResponseType: Optional[str] = None
    From: Optional[str] = None
    Subject: Optional[str] = None
    Body: Optional[str] = None
    ContentType: Optional[str] = None
    MimeType: Optional[str] = None
    Charset: Optional[str] = None
    CreateTime: Optional[str] = None
    ChangeTime: Optional[str] = None
    IncomingTime: Optional[int] = None
    To: Optional[str] = None
    SenderType: Optional[str] = None
    IsEdited: Optional[int] = None
    Cc: Optional[str] = None
    Bcc: Optional[str] = None
    ReplyTo: Optional[str] = None
    InReplyTo: Optional[str] = None
    References: Optional[str] = None
    MessageID: Optional[str] = None
    ContentCharset: Optional[str] = None
    ChangeBy: Optional[int] = None
    CreateBy: Optional[int] = None
    IsDeleted: Optional[int] = None
    ArticleID: Optional[int] = None
    ArticleNumber: Optional[int] = None
    DynamicField: Optional[List[DynamicFieldItem]] = None
    Attachment: Optional[AttachmentResponseDetail] = None


class TicketDetailOutput(TicketCommon):
    Article: Union[ArticleDetail, List[ArticleDetail]]
    DynamicField: List[DynamicFieldItem]

class TicketDetailInput(BaseModel):
    Ticket: Optional[TicketCommon] = None
    Article: Optional[Union[ArticleDetail, List[ArticleDetail]]] = None
    DynamicField: Optional[List[DynamicFieldItem]]= None
    Attachment: Optional[AttachmentModel] = None